/*Author : D Traore
 *
*/
/*
 * this files groups definition common to all adaptive algorithm
 */

#ifndef _TBB_ADAPT_H_
#define _TBB_ADAPT_H_

#include <algorithm>

#include <iostream>
#include <assert.h>
#include <vector>
#include <list>

#include <functional>

#include "tbb/task.h"
#include "tbb/mutex.h"
#include "tbb/atomic.h"

#define FINALIZE_SIGNAL

#define PTHREAD_MUTEX

namespace tbb_adapt {

inline uint log_2 ( uint a )  // returns [__log_2 a]+1
{
        uint res = 0 ;
        for ( ; a != 0 ; a >>=1, res++ ) ;
        return res ;
}

/*
 * forward declaration
 */
template<class T> class smart_ptr;
class WorkAdapt;
class FinalizeWork;
typedef smart_ptr<WorkAdapt*> WorkAdaptPtr;
typedef smart_ptr<FinalizeWork*> FinalizeWorkPtr;

/*
 * encapsulate a pointer and handles communication and memory allocation
 */

struct unmanaged {};

template <class T>
class smart_ptr<T*>
{
        T* _ptr;
        tbb::atomic<int> *_ref;

        void add_ref() { 
          _ref->fetch_and_increment();
        }

        void rm_ref(){
           if( _ref->fetch_and_decrement() == 0)
            {
                delete _ptr;
                delete _ref;
            }
        }

        friend class WorkAdapt;
        friend class FinalizeWork;

    public:
        smart_ptr():
            _ptr(0), _ref(0)
        {}

        smart_ptr(T* ptr):
            _ptr(ptr), _ref(new tbb::atomic<int>())
        {
            _ref->fetch_and_add(0);
            add_ref();
        }
        smart_ptr(T* ptr, unmanaged ):
            _ptr(ptr), _ref(0)
        {
        }
        
        smart_ptr(const smart_ptr<T*>& sp)
            :_ptr(sp._ptr), _ref(sp._ref)
        {
            if(_ref) add_ref();
        }

        smart_ptr<T*>& operator=(const smart_ptr<T*>& sp)
        {
            if( &sp != this)
            {
                if(_ref) rm_ref();
                _ptr=sp._ptr;
                _ref=sp._ref;
                if(_ref) add_ref();
            }
            return *this;
        }

        ~smart_ptr()
        {
            if( _ref) rm_ref();
        }

        T* operator->() { return _ptr; }
        const T* operator->() const { return _ptr; }
        T& operator*() { return *_ptr; }
        const T& operator*() const { return *_ptr; }
};


class NextJump {
  
   public :
     NextJump() { }
     virtual void init_work() const { }
     virtual void init_work() {}
};

class FinalizeWork {

    #ifdef PTHREAD_MUTEX
    pthread_mutex_t _mutex;
    #else 
    tbb::mutex _mutex;
    tbb::mutex::scoped_lock _s_lock;
    #endif


    #ifdef FINALIZE_SIGNAL
     bool _finalize_signal; 
    #endif

    protected:

        virtual bool extract_nextseq() = 0 ;
        virtual bool extract_par( FinalizeWork*& ) = 0 ;
        virtual void localcompute() =0 ;

   public :


     FinalizeWork() 
     {
        #ifdef PTHREAD_MUTEX
         pthread_mutex_init(&_mutex, NULL); 
        #endif
       
        #ifdef FINALIZE_SIGNAL
          _finalize_signal = false; 
       #endif
     }
     
     bool ExtractNextSeq()
     {
          #ifdef PTHREAD_MUTEX
           pthread_mutex_lock(&_mutex);
          #else
            _s_lock.acquire(_mutex);
          #endif
             bool local_extract_is_possible = extract_nextseq();
          #ifdef PTHREAD_MUTEX
           pthread_mutex_unlock(&_mutex);
          #else 
           _s_lock.release();
          #endif
         return local_extract_is_possible;
     }

     bool ExtractPar( FinalizeWorkPtr& wp)

     {
          #ifdef PTHREAD_MUTEX
            pthread_mutex_lock(&_mutex);
          #else
            _s_lock.acquire(_mutex);
          #endif
               bool steal_extract_is_possible = extract_par(wp._ptr);
          #ifdef PTHREAD_MUTEX 
           pthread_mutex_unlock(&_mutex);
          #else
            _s_lock.release();
          #endif
        return steal_extract_is_possible;
     }

     void LocalCompute() 
     {
       localcompute();
     }

   friend class WorkAdapt;

};


/*
 * main interface for adaptive works 
 */
class WorkAdapt
{
       #ifdef PTHREAD_MUTEX
        pthread_mutex_t _mutex;
       #else 
          tbb::mutex _mutex;
          tbb::mutex::scoped_lock _s_lock;
       #endif
        std::list< WorkAdapt* > thief_list;
        int not_finpar;
        bool _signalMerge;
        bool _signalJump;
        bool _progress;
        bool _update_work;
        #ifdef FINALIZE_SIGNAL
          bool _finalize_signal;
        #endif
        WorkAdapt* _thieft_work;
        FinalizeWork* _fw; 

    protected:

        virtual bool extract_nextseq() = 0 ;
        virtual bool extract_par( WorkAdapt/*Ptr*/*& ) = 0 ;
        virtual void localcompute() =0 ;
        virtual void join(const WorkAdapt* ) { };
        virtual bool extract_next_macrostep() { return false; };
        virtual void setjump(NextJump*& ) = 0;
        virtual void jump(const NextJump* )= 0;
        virtual void get_main_result(const WorkAdapt* ) {  };  
        virtual void get_thieft_result() {  };
        virtual void get_finalize_work(FinalizeWork*& ) {  };
        virtual inline bool is_finalize_work() const {  return false; }
        virtual bool is_empty() = 0 ;
 
    public:

        virtual inline bool  MustJump() const { return false; }
        virtual inline bool  MainJump() const { return true; } 

        WorkAdapt() 
        {
          #ifdef PTHREAD_MUTEX 
           pthread_mutex_init(&_mutex, NULL); 
          #endif
            not_finpar = 1;
           _signalMerge = false; 
           _signalJump = false; 
            _progress = true;
           _update_work = false;
           #ifdef FINALIZE_SIGNAL
             _finalize_signal = false;
           #endif

        }

        virtual ~WorkAdapt() {}

        bool ExtractNextSeq()
        {  
           #ifdef PTHREAD_MUTEX
             pthread_mutex_lock(&_mutex); 
           #else
            _s_lock.acquire(_mutex);
           #endif
            bool local_extract_is_possible = extract_nextseq();
            if(!local_extract_is_possible) not_finpar = 0;
          #ifdef PTHREAD_MUTEX
            pthread_mutex_unlock(&_mutex);
          #else
            _s_lock.release();  
          #endif
         return local_extract_is_possible; 
        }

        bool ExtractPar( WorkAdaptPtr& wp)
        {
          #ifdef PTHREAD_MUTEX
            pthread_mutex_lock(&_mutex); 
          #else
            _s_lock.acquire(_mutex);
          #endif
               bool steal_extract_is_possible = extract_par(wp._ptr);
               if(steal_extract_is_possible) thief_list.push_front(wp._ptr);
          #ifdef PTHREAD_MUTEX
            pthread_mutex_unlock(&_mutex);
          #else
            _s_lock.release();
          #endif
         return steal_extract_is_possible;
        }

       
        void LocalCompute()
        {
            localcompute();
        }

        void GetThieftWork( WorkAdapt*& thieftwork) {
           WorkAdapt *wa = thief_list.front();
           thieftwork = wa;
           thief_list.pop_front();
        }


        void Join(WorkAdapt*& jw )
        {
             join(jw);
        }
          
        void merge() {
             Join(_thieft_work);
        }


       bool getResultToPreempt(NextJump*& j)
       {
          bool go_to_next_thief = (nb_thief()!=0);
         #ifdef PTHREAD_MUTEX
          pthread_mutex_lock(&_mutex);
         #else
          _s_lock.acquire(_mutex);
         #endif
         if(_signalMerge) {
            go_to_next_thief = false;
            _signalJump = true;
             not_finpar = 0;
          } else  not_finpar = 1;
         #ifdef PTHREAD_MUTEX
          pthread_mutex_unlock(&_mutex);
         #else
          _s_lock.release();
         #endif
         if(go_to_next_thief) { 
             bool finalize_signal = false;
             return mergeWithNextThiefResult(j, finalize_signal) ; }
         else  {  not_finpar = 0; return false; }
      }

     bool mergeWithNextThiefResult(NextJump*& j, bool& finalize_signal)
      {   
         bool tmp;
         #ifdef PTHREAD_MUTEX
         pthread_mutex_lock(&_mutex);
         #else
         _s_lock.acquire(_mutex);
         #endif
            WorkAdapt *wa = thief_list.front();
            _thieft_work = wa;
            std::list< WorkAdapt* > tl ;
            if(is_finalize_work()) wa->get_main_result(this);
            wa->preemptAndGetResultToJump(j, finalize_signal);//preempted
            wa->getThiefList(tl);// Recuperation de la liste des voleurs de wa
            thief_list.splice(++thief_list.begin(), tl);
            #ifndef FINALIZE_SIGNAL
              _fw = wa->get_finalize();
            #else 
             _finalize_signal = finalize_signal; if(!_finalize_signal) _fw = wa->get_finalize();
            #endif

            thief_list.pop_front();

            tmp = _progress;

            if(!tmp) {
             _signalJump = true;
              not_finpar = 0;
             _update_work = true;
            merge(); //merge with thief result
            jump(j);//Jump to the next work
          }
       #ifdef PTHREAD_MUTEX
        pthread_mutex_unlock(&_mutex);
       #else
        _s_lock.release();
       #endif

      return tmp;
   }

      
      void preemptAndGetResultToJump(NextJump*& j, bool& finalize_signal)
      {
          bool tmp1, tmp2;
          #ifdef PTHREAD_MUTEX
           pthread_mutex_lock(&_mutex);
          #else
           _s_lock.acquire(_mutex);
          #endif
             _signalMerge = true;
              _progress = false;
             tmp1 = not_finpar;
             tmp2 = _signalJump;
             SetJump(j, finalize_signal);// permet d'arreter le calcul en cours
         #ifdef PTHREAD_MUTEX 
          pthread_mutex_unlock(&_mutex);
         #else
          _s_lock.release();
         #endif

          while(tmp1 && !tmp2){
           #ifdef PTHREAD_MUTEX
            pthread_mutex_lock(&_mutex);
           #else
            _s_lock.acquire(_mutex);
           #endif

              tmp1 = not_finpar;
              tmp2 = _signalJump;

           #ifdef PTHREAD_MUTEX 
            pthread_mutex_unlock(&_mutex);
           #else
            _s_lock.release();
           #endif
            sched_yield();
          }
 
          if(_update_work) {
            #ifdef PTHREAD_MUTEX
             pthread_mutex_lock(&_mutex);
            #else 
             _s_lock.acquire(_mutex);
            #endif
            
              SetJump(j, finalize_signal);
            
            #ifdef PTHREAD_MUTEX
             pthread_mutex_unlock(&_mutex);
            #else
             _s_lock.release();
            #endif
          }
      }


      void SetJump(NextJump*& j, bool& main_finalize_signal)
      {  
          if(is_finalize_work()) {
           
            get_finalize_work(_fw);
          
            #ifdef FINALIZE_SIGNAL 
            if(!is_empty()) {
              _finalize_signal = true;//For this
               main_finalize_signal = true; //For main process wich prempts
            }
            #endif

          }
          setjump(j);
      }
    
      void Jump(NextJump*& j){
        #ifdef PTHREAD_MUTEX
         pthread_mutex_lock(&_mutex);
        #else
         _s_lock.acquire(_mutex);
        #endif

          jump(j); //jump

        #ifdef PTHREAD_MUTEX
         pthread_mutex_unlock(&_mutex);
        #else
         _s_lock.release();
        #endif
      } 
   
      FinalizeWork* get_finalize() { return _fw;}
 
      void  getThiefList( std::list< WorkAdapt* >& tl)
      {
        #ifdef PTHREAD_MUTEX
          pthread_mutex_lock(&_mutex);
        #else
          _s_lock.acquire(_mutex);
        #endif

             tl = thief_list;

        #ifdef PTHREAD_MUTEX
          pthread_mutex_unlock(&_mutex);
        #else
         _s_lock.release();
        #endif
      }


      int nb_thief() { return thief_list.size(); }

     
      
      void set_signal_jump() {
          #ifdef PTHREAD_MUTEX
            pthread_mutex_lock(&_mutex);
          #else
           _s_lock.acquire(_mutex);
          #endif

             _signalJump = true;

          #ifdef PTHREAD_MUTEX
            pthread_mutex_unlock(&_mutex);
          #else
            _s_lock.release();
          #endif
      }

      void Run();

      class Main_Task: public tbb::task
      {

         WorkAdapt *current_work;

         public:

         Main_Task(WorkAdapt *_current_work): current_work(_current_work){ }

         tbb::task* execute();
     };

     friend class AdaptSteal;
};

/*
 * Adaptive engine for object inheriting from Work
 */

class FinalizeSteal: public tbb::task {

    FinalizeWorkPtr me;
    
    public:
 
    FinalizeSteal(FinalizeWorkPtr& _me): me(_me) { };
    

    tbb::task* execute()
    {
        FinalizeWorkPtr wp2;
         if(me->ExtractPar(wp2)) {
            
            FinalizeSteal& thief1 = *new(allocate_child() ) FinalizeSteal(me);
            FinalizeSteal& thief2 = *new(allocate_child() ) FinalizeSteal(wp2);
            set_ref_count(3);
            spawn(thief1);
            spawn(thief2);

            while(wp2->ExtractNextSeq())  wp2->LocalCompute();
            wait_for_all();//waiting
        }
      return NULL;
    }

};

/*
 Finalize work
*/
class Finalize: public tbb::task {

   FinalizeWorkPtr me;

   public:
   Finalize(FinalizeWorkPtr& _me):  me(_me) { }; 

   tbb::task* execute()
   {
         FinalizeSteal& thief1 = *new(allocate_child() ) FinalizeSteal(me);
         set_ref_count(2);
         spawn(thief1);
         while(me->ExtractNextSeq())  me->LocalCompute();
         wait_for_all();//waiting
     return NULL; 
   }
};

/*
 Steal work
*/
class AdaptSteal: public tbb::task 
{
    WorkAdaptPtr me;
  
    public: 
    AdaptSteal(WorkAdaptPtr& _me): me(_me) { };

    tbb::task* execute()
    {
        WorkAdaptPtr wp2;
         if(me->ExtractPar(wp2)) {
           AdaptSteal& thief1 = *new(allocate_child() ) AdaptSteal(me);
           set_ref_count(2);
           spawn(thief1); 
           bool tojump=true;
           int nbjump=0;
           while(tojump){
            AdaptSteal& thief2 =  *new(allocate_additional_child_of(*this)) AdaptSteal(wp2);
            spawn(thief2);
            while(wp2->ExtractNextSeq())  wp2->LocalCompute();
            if(wp2->MustJump()){
            NextJump *j ;
             if(wp2->getResultToPreempt(j)) {
               wp2->merge();
               wp2->Jump(j);
               //nbjump++; 
             } else { wp2->set_signal_jump(); tojump=false;}
            } else  { wp2->set_signal_jump(); tojump=false; }
          }//end microloop
         //if(nbjump)  std::cout << "thieft_nbjump=" << nbjump << std::endl;

        #ifdef FINALIZE_SIGNAL
         //Work To finalize after preemption
         if(wp2->_finalize_signal) {
              FinalizeWork *fw = wp2->get_finalize();
              FinalizeWorkPtr fw_ptr(fw);
              FinalizeSteal& thief3 = *new(allocate_additional_child_of(*this)) FinalizeSteal(fw_ptr);
              spawn(thief3);
              while(fw_ptr->ExtractNextSeq())  fw_ptr->LocalCompute();
         }
        #endif

         wait_for_all();//waiting
      }
    return NULL;
   }

};

 tbb::task* WorkAdapt::Main_Task::execute() {
    bool go_on ;
    bool no_spawn = true;
    int nbjump=0;
    do {
        go_on = current_work->extract_next_macrostep() ;
        {
          bool tojump=true;
          WorkAdaptPtr me ( current_work, unmanaged() );
          while(tojump){
            if(no_spawn) { 
             AdaptSteal& thief1 = *new(allocate_child() ) AdaptSteal(me);
             set_ref_count(2);
             spawn(thief1);
             no_spawn = false;
            } else {
              AdaptSteal& thief1 =  *new(allocate_additional_child_of(*this)) AdaptSteal(me);
              spawn(thief1);
            }
            //std::cout << "ref_count = " << ref_count() << std::endl;
            while( me->ExtractNextSeq() ) me->LocalCompute(); //nanoloop
            NextJump *j ;
            if(me->MainJump()){
              if(me->getResultToPreempt(j)) {
                if(me->is_finalize_work()){
                  #ifdef FINALIZE_SIGNAL
                   if(!(me->_finalize_signal)) {
                  #endif
                    FinalizeWork *fw = current_work->get_finalize();
                    FinalizeWorkPtr fw_ptr(fw);
                    Finalize& thief2 = *new(allocate_additional_child_of(*this) ) Finalize(fw_ptr);
                    spawn(thief2);
                  #ifdef FINALIZE_SIGNAL
                   }
                  #endif
                 }
                  me->merge();
                  me->Jump(j);
                  //nbjump++;
             } else tojump=false;
            } else tojump=false;
           }//end microloop
          //if(nbjump)   std::cout << "main_nbjump=" << nbjump << std::endl;
        }

    } while( go_on );
      
      //if(nbjump)   std::cout << "main_nbjump=" << nbjump << std::endl;
    
     wait_for_all();
     //std::cout << "End waiting...." << std::endl;
  return NULL;
 }


inline void WorkAdapt::Run()
{
  Main_Task& main_task = *new(tbb::task::allocate_root()) Main_Task(this); 
  tbb::task::spawn_root_and_wait(main_task);
}

inline void run_adapt(tbb_adapt::WorkAdapt* work)
{
  work->Run();   
};

} // en adapt

#endif // _TBB_ADAPT_H_
