#ifndef _TRANSFORM_H_
#define _TRANSFORM_H_

#include <numeric>
#include <cmath>
#include "../tbb_adapt.h"
#include "../jump_work.h"

namespace tbb_adapt {


///////////////////////////////////////UNARY TRANSFORM//////////////////////////////////////// 
template<class InputIterator, class OutputIterator, class UnaryOperator>
class UTransformWork : public tbb_adapt::WorkAdapt
{
    InputIterator _first;
    InputIterator _last;
    OutputIterator _res;
    UnaryOperator _u_op;
    int _beg;
    int _end;
    int _pargrain;
    int _c;
    int _seqgrain;
    int _beg_local;
    int _end_local;
    
    public:

     UTransformWork(InputIterator first, InputIterator last, OutputIterator res, UnaryOperator u_op,
         int beg, int c =1, int pargrain=2): WorkAdapt(), _first(first),  _last(last), _res(res), _u_op(u_op),
         _beg(beg), _end(last - first), _pargrain(pargrain), _c(c)
     {
           if(c <=0) _seqgrain = 1;
           else _seqgrain = _c * tbb_adapt::log_2(_end);
     }


    bool extract_nextseq()
    {
        
        if( _beg < _end ) {
            _beg_local = _beg; _beg +=_seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
           return true;
         } 
       return false;

    }


    bool extract_par( WorkAdapt*& sw)
    {
           if((_end - _beg) > _pargrain) {
             int i, j;
             int mid = _end - ((_end - _beg)/2);

             
             i = mid;
             j = _end;
             _end = mid;
             sw = new  UTransformWork<InputIterator, OutputIterator, UnaryOperator>(_first, _first+j, _res, _u_op, i, _c, _pargrain);
            return true;
           }
           else {
            sw = 0;
           return false;
          }
    }

    bool is_empty() {
      return (_beg >= _end);
    }

    void  localcompute()
    { 
       std::transform(_first + _beg_local, _first + _end_local, _res+_beg_local, _u_op) ;
    }

    //inline bool  MustJump() const { return true; }

    inline bool  MainJump() const { return false;}

    void setjump(tbb_adapt::NextJump*& j) {
        
          int end = _end;
          _end = _beg;

          j = new tbb_adapt::NextJumpWork(_beg, end);   
                
   }

   void jump(const tbb_adapt::NextJump* j) {
        
        const  tbb_adapt::NextJumpWork * nextj = dynamic_cast< const  tbb_adapt::NextJumpWork*> ( j );
        _beg = nextj->get_beg();
        _end = nextj->get_end();
   }

};

/////////////////////////////////////BINARY TRANSFORM//////////////////////////////////////////////
template < class InputIterator1, class InputIterator2,
           class OutputIterator, class BinaryOperator >
class BTransformWork : public tbb_adapt::WorkAdapt
{
    InputIterator1 _first1;
    InputIterator1 _last1;
    InputIterator2 _first2;
    OutputIterator _res;
    BinaryOperator _b_op;
    int _beg;
    int _end;
    int _pargrain;
    int _c;
    int _seqgrain;
    int _beg_local;
    int _end_local;

    public:

     BTransformWork(InputIterator1 first1, InputIterator1 last1, InputIterator2 first2,
          OutputIterator res, BinaryOperator b_op,
         int beg, int c =1, int pargrain=2): WorkAdapt(), _first1(first1),  _last1(last1), _first2(first2),
         _res(res), _b_op(b_op),
         _beg(beg), _end(last1 - first1), _pargrain(pargrain), _c(c)
     {
           if(c <=0) _seqgrain = 1;
           else _seqgrain = _c * tbb_adapt::log_2(_end);
     }


    bool extract_nextseq()
    {

        if( _beg < _end ) {
            _beg_local = _beg; _beg +=_seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
           return true;
         }
       return false;

    }

     bool extract_par( WorkAdapt*& sw)
    {
           if((_end - _beg) > _pargrain) {
             int i, j;
             int mid = _end - ((_end - _beg)/2);


             i = mid;
             j = _end;
             _end = mid;
             sw = new  BTransformWork<InputIterator1, InputIterator2, OutputIterator, BinaryOperator>(_first1, _first1+j,
                        _first2, _res, _b_op, i, _c, _pargrain);
            return true;
           }
           else {
            sw = 0;
           return false;
          }
    }

    bool is_empty() {
      return (_beg >= _end);
    }

    void  localcompute()
    {
       std::transform(_first1 + _beg_local, _first1 + _end_local, _first2+_beg_local, _res+_beg_local, _b_op) ;
    }

    //inline bool  MustJump() const { return true; }

    inline bool  MainJump() const { return false;}

   void setjump(tbb_adapt::NextJump*& j) {

          int end = _end;
          _end = _beg;

          j = new tbb_adapt::NextJumpWork(_beg, end);

   }

   void jump(const tbb_adapt::NextJump* j) {

        const  tbb_adapt::NextJumpWork * nextj = dynamic_cast< const  tbb_adapt::NextJumpWork*> ( j );
        _beg = nextj->get_beg();
        _end = nextj->get_end();
   }

};


////////////////////////////////////////////////////////////////////////////////////////////////

template <typename InputIterator, typename OutputIterator, typename UnaryOperator>
inline OutputIterator transform(InputIterator first, InputIterator last, OutputIterator res,
       UnaryOperator u_op, int seqgrain=100, int pargrain=2) 
{
    UTransformWork<InputIterator, OutputIterator, UnaryOperator>* work; 
    work = new  UTransformWork<InputIterator, OutputIterator, UnaryOperator>(first, last, res, u_op, 0, seqgrain);
    tbb_adapt::run_adapt(work);
  return res + (last-first);
}

template <typename InputIterator1, typename InputIterator2, typename OutputIterator, typename BinaryOperator>
inline OutputIterator transform(InputIterator1 first1, InputIterator1 last1, InputIterator2 first2, OutputIterator res, 
       BinaryOperator b_op, int seqgrain=100, int pargrain=2)
{
    BTransformWork<InputIterator1, InputIterator2, OutputIterator, BinaryOperator>* work;
    work = new  BTransformWork<InputIterator1, InputIterator2, OutputIterator, BinaryOperator>(first1, last1, 
                first2, res, b_op, 0, seqgrain);
    tbb_adapt::run_adapt(work);

   return res + (last1-first1);
}


}

#endif
